/*
 * File: Parser.java
 * -----------------
 * This class implements a very simple infix parser for expressions.
 */

import acm.util.*;
import java.io.*;
import java.util.*;

public class Parser {

/* Creates a new recursive-descent parser */
	public Parser() {
		symtab = new HashMap<String,Variable>();
	}

/* Sets the input for the parser to the specified input string */
	public void setInput(String input) {
		tokenizer = new StreamTokenizer(new StringReader(input));
		tokenizer.resetSyntax();
		tokenizer.wordChars('A', 'Z');
		tokenizer.wordChars('a', 'z');
		tokenizer.wordChars('0', '9');
		tokenizer.whitespaceChars(' ', ' ');
	}

/* Reads the next expression from the input stream */
	public Expression readExpression() {
		Expression exp = readE(0);
		Object token = readToken();
		if (token != null) {
			throw new ErrorException("Unexpected token " + token);
		}
		return exp;
	}

/*
 * Reads the next instance of the grammar category E, stopping
 * when an operator has a precedence less than the prevailing
 * precedence specified by the prec parameter.
 */
	private Expression readE(int prec) {
		Expression exp = readT();
		Object token = readToken();
		while (token != null) {
			char op = ((Character) token).charValue();
			int newPrec = precedence(op);
			if (prec >= newPrec) break;
			exp = createCompound(op, exp, readE(newPrec));
			token = readToken();
		}
		saveToken(token);
		return exp;
	}

/*
 * Reads the next instance of the grammar category T, which
 * represents an individual term.
 */
	private Expression readT() {
		Object token = readToken();
		if (token == null) {
			throw new ErrorException("Unexpected end of line");
		}
		if (token instanceof Character) {
			if (((Character) token).charValue() == '(') {
				Expression exp = readE(0);
				token = readToken();
				if (!(token instanceof Character) || ((Character) token).charValue() != ')') {
					throw new ErrorException("Mismatched parentheses");
				}
				return exp;
			}
			throw new ErrorException("Unrecognized operator " + token);
		}
		return (Expression) token;
	}

/* Returns the precedence of the specified operator character */
	private int precedence(char op) {
		switch (op) {
		  case '*': case '/': case '%': return 90;
		  case '+': case '-': return 80;
		  case '=': return 10;
		  case ')': return 0;
		  default: throw new ErrorException("Unrecognized operator " + op);
		}
	}

/* Creates a new compound from the specified parts */
	private Compound createCompound(char op, Expression lhs, Expression rhs) {
		switch (op) {
		  case '*': return new MultiplyCompound(op, lhs, rhs);
		  case '/': return new DivideCompound(op, lhs, rhs);
		  case '%': return new RemainderCompound(op, lhs, rhs);
		  case '+': return new AddCompound(op, lhs, rhs);
		  case '-': return new SubtractCompound(op, lhs, rhs);
		  case '=': return new AssignmentCompound(op, lhs, rhs);
		  default: throw new ErrorException("Unrecognized operator " + op);
		}
	}

/*
 * Reads a single token from the input stream.  The tokens are always
 * one of the following types:
 *
 * <ul>
 * <li>A <code>Constant</code>, representing an integer constant
 * <li>A <code>Variable</code>, representing a variable name
 * <li>A <code>Character</code>, representing any operator character
 * </ul>
 */
	private Object readToken() {
		if (isTokenSaved) {
			isTokenSaved = false;
			return savedToken;
		}
		try {
			int tt = tokenizer.nextToken();
			if (tt == StreamTokenizer.TT_EOF) return null;
			if (tt == StreamTokenizer.TT_WORD) {
				if (Character.isDigit(tokenizer.sval.charAt(0))) {
					return new Constant(Integer.parseInt(tokenizer.sval));
				} else {
					Variable var = (Variable) symtab.get(tokenizer.sval);
					if (var == null) {
						var = new Variable(tokenizer.sval);
						symtab.put(tokenizer.sval, var);
					}
					return var;
				}
			}
			return new Character((char) tt);
		} catch (Exception ex) {
			throw new ErrorException("Error: " + ex);
		}
	}

/*
 * Saves the specified token so that it will be returned on
 * the next call to <code>readToken</code>.
 */
	private void saveToken(Object token) {
		isTokenSaved = true;
		savedToken = token;
	}

/* Private instance variables */
	private HashMap<String,Variable> symtab;
	private StreamTokenizer tokenizer;
	private boolean isTokenSaved;
	private Object savedToken;

}

class MultiplyCompound extends Compound {
	public MultiplyCompound(char op, Expression lhs, Expression rhs) {
		super(op, lhs, rhs);
	}

	public int apply(int lhs, int rhs) {
		return lhs * rhs;
	}
}

class DivideCompound extends Compound {
	public DivideCompound(char op, Expression lhs, Expression rhs) {
		super(op, lhs, rhs);
	}

	public int apply(int lhs, int rhs) {
		return lhs / rhs;
	}
}

class RemainderCompound extends Compound {
	public RemainderCompound(char op, Expression lhs, Expression rhs) {
		super(op, lhs, rhs);
	}

	public int apply(int lhs, int rhs) {
		return lhs % rhs;
	}
}

class AddCompound extends Compound {
	public AddCompound(char op, Expression lhs, Expression rhs) {
		super(op, lhs, rhs);
	}

	public int apply(int lhs, int rhs) {
		return lhs + rhs;
	}
}

class SubtractCompound extends Compound {
	public SubtractCompound(char op, Expression lhs, Expression rhs) {
		super(op, lhs, rhs);
	}

	public int apply(int lhs, int rhs) {
		return lhs - rhs;
	}
}

class AssignmentCompound extends Compound {
	public AssignmentCompound(char op, Expression lhs, Expression rhs) {
		super(op, lhs, rhs);
		if (!(lhs instanceof Variable)) {
			throw new ErrorException("Illegal assignment");
		}
		this.lhs = (Variable) lhs;
		this.rhs = rhs;
	}

	public int eval() {
		int value = rhs.eval();
		lhs.assignValue(value);
		return value;
	}

	public int apply(int lhs, int rhs) {
		throw new ErrorException("Never called");
	}

	private Variable lhs;
	private Expression rhs;
}
